@@ -9,7 +9,7 @@
 ***********************************************************************/
 
 static char const RCSID[] =
-"$Id: discovery.c,v 1.10 2001/01/29 17:37:20 dfs Exp $";
+"$Id: discovery.c,v 1.15 2001/07/13 13:22:16 dfs Exp $";
 
 #include "pppoe.h"
 
@@ -82,6 +82,9 @@ packetIsForMe(PPPoEConnection *conn, PPPoEPacket *packet)
 {
     int forMe = 0;
 
+    /* If packet is not directed to our MAC address, forget it */
+    if (memcmp(packet->ethHdr.h_dest, conn->myEth, ETH_ALEN)) return 0;
+
     /* If we're not using the Host-Unique tag, then accept the packet */
     if (!conn->useHostUniq) return 1;
 
@@ -113,6 +116,7 @@ parsePADOTags(UINT16_t type, UINT16_t len, unsigned char *data,
 
     switch(type) {
     case TAG_AC_NAME:
+	pc->seenACName = 1;
 	if (conn->printACNames) {
 	    printf("Access-Concentrator: %.*s\n", (int) len, data);
 	}
@@ -122,6 +126,7 @@ parsePADOTags(UINT16_t type, UINT16_t len, unsigned char *data,
 	}
 	break;
     case TAG_SERVICE_NAME:
+	pc->seenServiceName = 1;
 	if (conn->printACNames && len > 0) {
 	    printf("       Service-Name: %.*s\n", (int) len, data);
 	}
@@ -284,10 +289,9 @@ sendPADI(PPPoEConnection *conn)
 
     packet.length = htons(plen);
 
-    sendPacket(conn->discoverySocket, &packet, (int) (plen + HDR_SIZE));
+    sendPacket(conn, conn->discoverySocket, &packet, (int) (plen + HDR_SIZE));
     if (conn->debugFile) {
-	fprintf(conn->debugFile, "SENT ");
-	dumpPacket(conn->debugFile, &packet);
+	dumpPacket(conn->debugFile, &packet, "SENT");
 	fprintf(conn->debugFile, "\n");
 	fflush(conn->debugFile);
     }
@@ -316,6 +320,8 @@ waitForPADO(PPPoEConnection *conn, int timeout)
     pc.conn          = conn;
     pc.acNameOK      = (conn->acName)      ? 0 : 1;
     pc.serviceNameOK = (conn->serviceName) ? 0 : 1;
+    pc.seenACName    = 0;
+    pc.seenServiceName = 0;
 	
     do {
 	if (BPF_BUFFER_IS_EMPTY) {
@@ -351,8 +357,7 @@ waitForPADO(PPPoEConnection *conn, int timeout)
 #endif
 
 	if (conn->debugFile) {
-	    fprintf(conn->debugFile, "RCVD ");
-	    dumpPacket(conn->debugFile, &packet);
+	    dumpPacket(conn->debugFile, &packet, "RCVD");
 	    fprintf(conn->debugFile, "\n");
 	    fflush(conn->debugFile);
 	}
@@ -364,11 +369,19 @@ waitForPADO(PPPoEConnection *conn, int timeout)
 		printErr("Ignoring PADO packet from non-unicast MAC address");
 		continue;
 	    }
+	    parsePacket(&packet, parsePADOTags, &pc);
+	    if (!pc.seenACName) {
+		printErr("Ignoring PADO packet with no AC-Name tag");
+		continue;
+	    }
+	    if (!pc.seenServiceName) {
+		printErr("Ignoring PADO packet with no Service-Name tag");
+		continue;
+	    }
 	    conn->numPADOs++;
 	    if (conn->printACNames) {
 		printf("--------------------------------------------------\n");
 	    }
-	    parsePacket(&packet, parsePADOTags, &pc);
 	    if (pc.acNameOK && pc.serviceNameOK) {
 		memcpy(conn->peerEth, packet.ethHdr.h_source, ETH_ALEN);
 		if (conn->printACNames) {
@@ -460,10 +473,9 @@ sendPADR(PPPoEConnection *conn)
     }
 
     packet.length = htons(plen);
-    sendPacket(conn->discoverySocket, &packet, (int) (plen + HDR_SIZE));
+    sendPacket(conn, conn->discoverySocket, &packet, (int) (plen + HDR_SIZE));
     if (conn->debugFile) {
-	fprintf(conn->debugFile, "SENT ");
-	dumpPacket(conn->debugFile, &packet);
+	dumpPacket(conn->debugFile, &packet, "SENT");
 	fprintf(conn->debugFile, "\n");
 	fflush(conn->debugFile);
     }
@@ -521,8 +533,7 @@ waitForPADS(PPPoEConnection *conn, int timeout)
 	if (etherType(&packet) != Eth_PPPOE_Discovery) continue;
 #endif
 	if (conn->debugFile) {
-	    fprintf(conn->debugFile, "RCVD ");
-	    dumpPacket(conn->debugFile, &packet);
+	    dumpPacket(conn->debugFile, &packet, "RCVD");
 	    fprintf(conn->debugFile, "\n");
 	    fflush(conn->debugFile);
 	}
@@ -569,6 +580,12 @@ discovery(PPPoEConnection *conn)
     int padrAttempts = 0;
     int timeout = PADI_TIMEOUT;
 
+    /* Skip discovery and don't open discovery socket? */
+    if (conn->skipDiscovery && conn->noDiscoverySocket) {
+	conn->discoveryState = STATE_SESSION;
+	return;
+    }
+
     conn->discoverySocket =
 	openInterface(conn->ifName, Eth_PPPOE_Discovery, conn->myEth);
 
