@@ -14,7 +14,7 @@
 ***********************************************************************/
 
 static char const RCSID[] =
-"$Id: if.c,v 1.5 2001/01/26 18:15:12 dfs Exp $";
+"$Id: if.c,v 1.12 2001/08/06 01:57:19 dfs Exp $";
 
 #include "pppoe.h"
 
@@ -48,6 +48,10 @@ static char const RCSID[] =
 #include <stdlib.h>
 #include <string.h>
 
+#ifdef HAVE_NET_IF_ARP_H
+#include <net/if_arp.h>
+#endif
+
 #ifdef USE_DLPI
 
 #include <limits.h>
@@ -97,8 +101,6 @@ int bpfSize = 0;		/* Number of unread bytes in buffer */
 int bpfOffset = 0;		/* Current offset in bpfBuffer */
 #endif
 
-extern char *IfName;	/* For brain-dead Linux 2.0-kernel SOCK_PACKET */
-
 /* Initialize frame types to RFC 2516 values.  Some broken peers apparently
    use different frame types... sigh... */
 
@@ -204,35 +206,44 @@ initFilter(int fd, UINT16_t type, unsigned char *hwaddr)
     /* Packet Filter Instructions:
      * Note that the ethernet type names come from "pppoe.h" and are
      * used here to maintain consistency with the rest of this file. */
-    static struct bpf_insn bpfRun[] = {		/* run PPPoE */
-	BPF_STMT(BPF_LD+BPF_H+BPF_ABS, 12),	/* ethernet type */
-	BPF_JUMP(BPF_JMP+BPF_JEQ+BPF_K, ETH_PPPOE_SESSION, 1, 0),
-	BPF_JUMP(BPF_JMP+BPF_JEQ+BPF_K, ETH_PPPOE_DISCOVERY, 0, 5),
-	BPF_STMT(BPF_LD+BPF_W+BPF_ABS, 0),	/* first word of dest. addr */
-#define PPPOE_FILTER_CMPW 4			/* offset of word compare */
-	BPF_JUMP(BPF_JMP+BPF_JEQ+BPF_K, 0, 0, 3),
-	BPF_STMT(BPF_LD+BPF_H+BPF_ABS, 4),	/* next 1/2 word of dest. */
-#define PPPOE_FILTER_CMPH 6			/* offset of 1/2 word compare */
-	BPF_JUMP(BPF_JMP+BPF_JEQ+BPF_K, 0, 0, 1),
-	BPF_STMT(BPF_RET+BPF_K, (u_int) -1),	/* keep packet */
-	BPF_STMT(BPF_RET+BPF_K, 0),		/* drop packet */
+    static struct bpf_insn bpfRun[] = {         /* run PPPoE */
+        BPF_STMT(BPF_LD+BPF_H+BPF_ABS, 12),     /* ethernet type */
+        BPF_JUMP(BPF_JMP+BPF_JEQ+BPF_K, ETH_PPPOE_SESSION, 5, 0),
+        BPF_JUMP(BPF_JMP+BPF_JEQ+BPF_K, ETH_PPPOE_DISCOVERY, 0, 9),
+        BPF_STMT(BPF_LD+BPF_W+BPF_ABS, 0),      /* first word of dest. addr */
+#define PPPOE_BCAST_CMPW 4                     /* offset of word compare */
+        BPF_JUMP(BPF_JMP+BPF_JEQ+BPF_K, 0, 0, 2),
+        BPF_STMT(BPF_LD+BPF_H+BPF_ABS, 4),      /* next 1/2 word of dest. */
+#define PPPOE_BCAST_CMPH 6                     /* offset of 1/2 word compare */
+        BPF_JUMP(BPF_JMP+BPF_JEQ+BPF_K, 0, 4, 0),
+        BPF_STMT(BPF_LD+BPF_W+BPF_ABS, 0),      /* first word of dest. addr */
+#define PPPOE_FILTER_CMPW 8                     /* offset of word compare */
+        BPF_JUMP(BPF_JMP+BPF_JEQ+BPF_K, 0, 0, 3),
+        BPF_STMT(BPF_LD+BPF_H+BPF_ABS, 4),      /* next 1/2 word of dest. */
+#define PPPOE_FILTER_CMPH 10                    /* offset of 1/rd compare */
+        BPF_JUMP(BPF_JMP+BPF_JEQ+BPF_K, 0, 0, 1),
+        BPF_STMT(BPF_RET+BPF_K, (u_int) -1),    /* keep packet */
+        BPF_STMT(BPF_RET+BPF_K, 0),             /* drop packet */
     };
 
     /* Fix the potentially varying parts */
     bpfRun[1].code = (u_short) BPF_JMP+BPF_JEQ+BPF_K;
-    bpfRun[1].jt   = 1;
+    bpfRun[1].jt   = 5;
     bpfRun[1].jf   = 0;
     bpfRun[1].k    = Eth_PPPOE_Session;
 
     bpfRun[2].code = (u_short) BPF_JMP+BPF_JEQ+BPF_K;
     bpfRun[2].jt   = 0;
-    bpfRun[2].jf   = 5;
+    bpfRun[2].jf   = 9;
     bpfRun[2].k    = Eth_PPPOE_Discovery;
 
     {
       struct bpf_insn bpfInsn[sizeof(bpfRun) / sizeof(bpfRun[0])];
       struct bpf_program bpfProgram;
       memcpy(bpfInsn, bpfRun, sizeof(bpfRun));
+      bpfInsn[PPPOE_BCAST_CMPW].k = ((0xff << 24) | (0xff << 16) |
+                                     (0xff << 8) | 0xff);
+      bpfInsn[PPPOE_BCAST_CMPH].k = ((0xff << 8) | 0xff);
       bpfInsn[PPPOE_FILTER_CMPW].k = ((hwaddr[0] << 24) | (hwaddr[1] << 16) |
 				      (hwaddr[2] << 8) | hwaddr[3]);
       bpfInsn[PPPOE_FILTER_CMPH].k = ((hwaddr[4] << 8) | hwaddr[5]);
@@ -447,6 +458,13 @@ openInterface(char const *ifname, UINT16_t type, unsigned char *hwaddr)
 	    fatalSys("ioctl(SIOCGIFHWADDR)");
 	}
 	memcpy(hwaddr, ifr.ifr_hwaddr.sa_data, ETH_ALEN);
+#ifdef ARPHRD_ETHER
+	if (ifr.ifr_hwaddr.sa_family != ARPHRD_ETHER) {
+	    char buffer[256];
+	    sprintf(buffer, "Interface %.16s is not Ethernet", ifname);
+	    rp_fatal(buffer);
+	}
+#endif
 	if (NOT_UNICAST(hwaddr)) {
 	    char buffer[256];
 	    sprintf(buffer,
@@ -500,20 +518,22 @@ openInterface(char const *ifname, UINT16_t type, unsigned char *hwaddr)
 * pkt -- the packet to transmit
 * size -- size of packet (in bytes)
 *%RETURNS:
-* Nothing
+* 0 on success; -1 on failure
 *%DESCRIPTION:
 * Transmits a packet
 ***********************************************************************/
-void
-sendPacket(int sock, PPPoEPacket *pkt, int size)
+int
+sendPacket(PPPoEConnection *conn, int sock, PPPoEPacket *pkt, int size)
 {
 #if defined(USE_BPF)
     if (write(sock, pkt, size) < 0) {
-	fatalSys("write (sendPacket)");
+	sysErr("write (sendPacket)");
+	return -1;
     }
 #elif defined(HAVE_STRUCT_SOCKADDR_LL)
     if (send(sock, pkt, size, 0) < 0) {
-	fatalSys("send (sendPacket)");
+	sysErr("send (sendPacket)");
+	return -1;
     }
 #else
 #ifdef USE_DLPI
@@ -555,12 +575,18 @@ sendPacket(int sock, PPPoEPacket *pkt, int size)
 
 #else
     struct sockaddr sa;
-    strcpy(sa.sa_data, IfName);
+
+    if (!conn) {
+	rp_fatal("relay and server not supported on Linux 2.0 kernels");
+    }
+    strcpy(sa.sa_data, conn->ifName);
     if (sendto(sock, pkt, size, 0, &sa, sizeof(sa)) < 0) {
-	fatalSys("sendto (sendPacket)");
+	sysErr("sendto (sendPacket)");
+	return -1;
     }
 #endif
 #endif
+    return 0;
 }
 
 #ifdef USE_BPF
@@ -697,6 +723,9 @@ openInterface(char const *ifname, UINT16_t type, unsigned char *hwaddr)
     strncpy(base_dev, ifname, PATH_MAX); 
     base_dev[strlen(base_dev)-1] = '\0'; 
 
+/* rearranged order of DLPI code - delphys 20010803 */
+    dlp = (union DL_primitives*) buf;
+
     if (( fd = open(base_dev, O_RDWR)) < 0) {
 	/* Give a more helpful message for the common error case */
 	if (errno == EPERM) {
@@ -705,28 +734,30 @@ openInterface(char const *ifname, UINT16_t type, unsigned char *hwaddr)
 	fatalSys("socket");
     }
 
-	dlinforeq(fd);
-	dlinfoack(fd, (char *)buf);
+/* rearranged order of DLPI code - delphys 20010803 */
+    dlattachreq(fd, ppa); 
+    dlokack(fd, (char *)buf);
 
-	dlp = (union DL_primitives*) buf;
+    dlbindreq(fd, type, 0, DL_CLDLS, 0, 0);
+    dlbindack(fd, (char *)buf);
 
-	dl_abssaplen = ABS(dlp->info_ack.dl_sap_length);
-	dl_saplen = dlp->info_ack.dl_sap_length;
-	if (ETHERADDRL != (dlp->info_ack.dl_addr_length - dl_abssaplen))
-		fatalSys("invalid destination physical address length");
-	dl_addrlen = dl_abssaplen + ETHERADDRL;
+    dlinforeq(fd);
+    dlinfoack(fd, (char *)buf);
 
-	dlattachreq(fd, ppa); 
-	dlokack(fd, (char *)buf);
+    dl_abssaplen = ABS(dlp->info_ack.dl_sap_length);
+    dl_saplen = dlp->info_ack.dl_sap_length;
+    if (ETHERADDRL != (dlp->info_ack.dl_addr_length - dl_abssaplen))
+	fatalSys("invalid destination physical address length");
+    dl_addrlen = dl_abssaplen + ETHERADDRL;
 
-	dlbindreq(fd, type, 0, DL_CLDLS, 0, 0);
-	dlbindack(fd, (char *)buf);
+/* ethernet address retrieved as part of DL_INFO_ACK - delphys 20010803 */
+    memcpy(hwaddr, (u_char*)((char*)(dlp) + (int)(dlp->info_ack.dl_addr_offset)), ETHERADDRL);
 
-	if ( strioctl(fd, DLIOCRAW, -1, 0, NULL) < 0 ) { 
-		fatalSys("DLIOCRAW"); 
-	}
+    if ( strioctl(fd, DLIOCRAW, -1, 0, NULL) < 0 ) { 
+	fatalSys("DLIOCRAW"); 
+    }
 
-	if (ioctl(fd, I_FLUSH, FLUSHR) < 0) fatalSys("I_FLUSH");
+    if (ioctl(fd, I_FLUSH, FLUSHR) < 0) fatalSys("I_FLUSH");
 
     return fd;
 }
