@@ -12,7 +12,7 @@
 ***********************************************************************/
 
 static char const RCSID[] =
-"$Id: pppoe.c,v 1.12 2001/01/29 17:37:20 dfs Exp $";
+"$Id: pppoe.c,v 1.27 2001/09/03 20:09:19 dfs Exp $";
 
 #include "pppoe.h"
 
@@ -61,6 +61,9 @@ int optInactivityTimeout = 0;	/* Inactivity timeout */
 int optClampMSS          = 0;	/* Clamp MSS to this value */
 int optSkipSession       = 0;	/* Perform discovery, print session info
 				   and exit */
+int optFloodDiscovery    = 0;   /* Flood server with discovery requests.
+				   USED FOR STRESS-TESTING ONLY.  DO NOT
+				   USE THE -F OPTION AGAINST A REAL ISP */
 
 PPPoEConnection *Connection = NULL; /* Must be global -- used
 				       in signal handler */
@@ -82,10 +85,11 @@ sendSessionPacket(PPPoEConnection *conn, PPPoEPacket *packet, int len)
     if (optClampMSS) {
 	clampMSS(packet, "outgoing", optClampMSS);
     }
-    sendPacket(conn->sessionSocket, packet, len + HDR_SIZE);
+    if (sendPacket(conn, conn->sessionSocket, packet, len + HDR_SIZE) < 0) {
+	exit(EXIT_FAILURE);
+    }
     if (conn->debugFile) {
-	fprintf(conn->debugFile, "SENT ");
-	dumpPacket(conn->debugFile, packet);
+	dumpPacket(conn->debugFile, packet, "SENT");
 	fprintf(conn->debugFile, "\n");
 	fflush(conn->debugFile);
     }
@@ -117,13 +121,20 @@ sessionDiscoveryPacket(PPPoEPacket *packet)
     }
 
     /* It's a PADT, all right.  Is it for us? */
-    if (packet->session != Session) {
+    if (packet->session != Connection->session) {
 	/* Nope, ignore it */
 	return;
     }
+    if (memcmp(packet->ethHdr.h_dest, Connection->myEth, ETH_ALEN)) {
+	return;
+    }
+
+    if (memcmp(packet->ethHdr.h_source, Connection->peerEth, ETH_ALEN)) {
+	return;
+    }
 
     syslog(LOG_INFO,
-	   "Session terminated -- received PADT from access concentrator");
+	   "Session terminated -- received PADT from peer");
     parsePacket(packet, parseLogErrs, NULL);
     exit(EXIT_SUCCESS);
 }
@@ -155,13 +166,6 @@ sessionDiscoveryPacket(PPPoEConnection *conn)
 	return;
     }
 
-    if (conn->debugFile) {
-	fprintf(conn->debugFile, "RCVD ");
-	dumpPacket(conn->debugFile, &packet);
-	fprintf(conn->debugFile, "\n");
-	fflush(conn->debugFile);
-    }
-
     if (packet.code != CODE_PADT) {
 	/* Not PADT; ignore it */
 	return;
@@ -173,6 +177,20 @@ sessionDiscoveryPacket(PPPoEConnection *conn)
 	return;
     }
 
+    if (memcmp(packet.ethHdr.h_dest, conn->myEth, ETH_ALEN)) {
+	return;
+    }
+
+    if (memcmp(packet.ethHdr.h_source, conn->peerEth, ETH_ALEN)) {
+	return;
+    }
+
+    if (conn->debugFile) {
+	dumpPacket(conn->debugFile, &packet, "RCVD");
+	fprintf(conn->debugFile, "\n");
+	fflush(conn->debugFile);
+    }
+
     syslog(LOG_INFO,
 	   "Session terminated -- received PADT from peer");
     parsePacket(&packet, parseLogErrs, NULL);
@@ -222,9 +240,9 @@ session(PPPoEConnection *conn)
     /* check for buffered session data */
     while (BPF_BUFFER_HAS_DATA) {
 	if (conn->synchronous) {
-	    syncReadFromEth(conn->sessionSocket, optClampMSS);
+	    syncReadFromEth(conn, conn->sessionSocket, optClampMSS);
 	} else {
-	    asyncReadFromEth(conn->sessionSocket, optClampMSS);
+	    asyncReadFromEth(conn, conn->sessionSocket, optClampMSS);
 	}
     }
 #endif
@@ -237,7 +255,9 @@ session(PPPoEConnection *conn)
 	}
 	FD_ZERO(&readable);
 	FD_SET(0, &readable);     /* ppp packets come from stdin */
-	FD_SET(conn->discoverySocket, &readable);
+	if (conn->discoverySocket >= 0) {
+	    FD_SET(conn->discoverySocket, &readable);
+	}
 	FD_SET(conn->sessionSocket, &readable);
 	while(1) {
 	    r = select(maxFD, &readable, NULL, NULL, tvp);
@@ -274,8 +294,10 @@ session(PPPoEConnection *conn)
 #ifndef USE_BPF	
 	/* BSD uses a single socket, see *syncReadFromEth() */
 	/* for calls to sessionDiscoveryPacket() */
-	if (FD_ISSET(conn->discoverySocket, &readable)) {
-	    sessionDiscoveryPacket(conn);
+	if (conn->discoverySocket >= 0) {
+	    if (FD_ISSET(conn->discoverySocket, &readable)) {
+		sessionDiscoveryPacket(conn);
+	    }
 	}
 #endif
 
@@ -332,6 +354,7 @@ usage(char const *argv0)
     fprintf(stderr, "   -m MSS         -- Clamp incoming and outgoing MSS options.\n");
     fprintf(stderr, "   -p pidfile     -- Write process-ID to pidfile.\n");
     fprintf(stderr, "   -e sess:mac    -- Skip discovery phase; use existing session.\n");
+    fprintf(stderr, "   -n             -- Do not open discovery socket.\n");
     fprintf(stderr, "   -k             -- Kill a session with PADT (requires -e)\n");
     fprintf(stderr, "   -d             -- Perform discovery, print session info and exit.\n");
     fprintf(stderr, "   -f disc:sess   -- Set Ethernet frame types (hex).\n");
@@ -381,8 +404,19 @@ main(int argc, char *argv[])
     /* Initialize syslog */
     openlog("pppoe", LOG_PID, LOG_DAEMON);
 
-    while((opt = getopt(argc, argv, "I:VAT:D:hS:C:Usm:p:e:kdf:")) != -1) {
+    while((opt = getopt(argc, argv, "I:VAT:D:hS:C:Usm:np:e:kdf:F:")) != -1) {
 	switch(opt) {
+	case 'F':
+	    if (sscanf(optarg, "%d", &optFloodDiscovery) != 1) {
+		fprintf(stderr, "Illegal argument to -F: Should be -F numFloods\n");
+		exit(EXIT_FAILURE);
+	    }
+	    if (optFloodDiscovery < 1) optFloodDiscovery = 1;
+	    fprintf(stderr, "WARNING: DISCOVERY FLOOD IS MEANT FOR STRESS-TESTING\n");
+	    fprintf(stderr, "A PPPOE SERVER WHICH YOU OWN.  DO NOT USE IT AGAINST\n");
+	    fprintf(stderr, "A REAL ISP.  YOU HAVE 5 SECONDS TO ABORT.\n");
+	    sleep(5);
+	    break;
 	case 'f':
 	    if (sscanf(optarg, "%x:%x", &discoveryType, &sessionType) != 2) {
 		fprintf(stderr, "Illegal argument to -f: Should be disc:sess in hex\n");
@@ -399,6 +433,12 @@ main(int argc, char *argv[])
 	    conn.killSession = 1;
 	    break;
 
+	case 'n':
+	    /* Do not even open a discovery socket -- used when invoked 
+	       by pppoe-server */
+	    conn.noDiscoverySocket = 1;
+	    break;
+
 	case 'e':
 	    /* Existing session: "sess:xx:yy:zz:aa:bb:cc" where "sess" is
 	       session-ID, and xx:yy:zz:aa:bb:cc is MAC-address of peer */
@@ -523,6 +563,18 @@ main(int argc, char *argv[])
 
     }
 
+    if (optFloodDiscovery) {
+	for (n=0; n < optFloodDiscovery; n++) {
+	    if (conn.printACNames) {
+		fprintf(stderr, "Sending discovery flood %d\n", n+1);
+	    }
+	    discovery(&conn);
+	    conn.discoveryState = STATE_SENT_PADI;
+	    close(conn.discoverySocket);
+	}
+	exit(EXIT_SUCCESS);
+    }
+
     discovery(&conn);
     if (optSkipSession) {
 	printf("%u:%02x:%02x:%02x:%02x:%02x:%02x\n",
@@ -587,8 +639,10 @@ sysErr(char const *str)
 void
 rp_fatal(char const *str)
 {
+    char buf[1024];
     printErr(str);
-    sendPADT(Connection, str);
+    sprintf(buf, "RP-PPPoE: %.256s", str);
+    sendPADT(Connection, buf);
     exit(EXIT_FAILURE);
 }
 
@@ -632,8 +686,7 @@ asyncReadFromEth(PPPoEConnection *conn, int sock, int clampMss)
 	return;
     }
     if (conn->debugFile) {
-	fprintf(conn->debugFile, "RCVD ");
-	dumpPacket(conn->debugFile, &packet);
+	dumpPacket(conn->debugFile, &packet, "RCVD");
 	fprintf(conn->debugFile, "\n");
 	fflush(conn->debugFile);
     }
@@ -661,6 +714,9 @@ asyncReadFromEth(PPPoEConnection *conn, int sock, int clampMss)
 	syslog(LOG_ERR, "Unexpected packet type %d", (int) packet.type);
 	return;
     }
+    if (memcmp(packet.ethHdr.h_dest, conn->myEth, ETH_ALEN)) {
+	return;
+    }
     if (memcmp(packet.ethHdr.h_source, conn->peerEth, ETH_ALEN)) {
 	/* Not for us -- must be another session.  This is not an error,
 	   so don't log anything.  */
@@ -757,8 +813,7 @@ syncReadFromEth(PPPoEConnection *conn, int sock, int clampMss)
 	return;
     }
     if (conn->debugFile) {
-	fprintf(conn->debugFile, "RCVD ");
-	dumpPacket(conn->debugFile, &packet);
+	dumpPacket(conn->debugFile, &packet, "RCVD");
 	fprintf(conn->debugFile, "\n");
 	fflush(conn->debugFile);
     }
@@ -786,6 +841,11 @@ syncReadFromEth(PPPoEConnection *conn, int sock, int clampMss)
 	syslog(LOG_ERR, "Unexpected packet type %d", (int) packet.type);
 	return;
     }
+    if (memcmp(packet.ethHdr.h_dest, conn->myEth, ETH_ALEN)) {
+	/* Not for us -- must be another session.  This is not an error,
+	   so don't log anything.  */
+	return;
+    }
     if (memcmp(packet.ethHdr.h_source, conn->peerEth, ETH_ALEN)) {
 	/* Not for us -- must be another session.  This is not an error,
 	   so don't log anything.  */
