@@ -9,11 +9,11 @@
 * This program may be distributed according to the terms of the GNU
 * General Public License, version 2 or (at your option) any later version.
 *
-* $Id: relay.c,v 1.19 2001/02/05 02:20:53 dfs Exp $
+* $Id: relay.c,v 1.22 2001/06/14 20:59:20 dfs Exp $
 *
 ***********************************************************************/
 static char const RCSID[] =
-"$Id: relay.c,v 1.19 2001/02/05 02:20:53 dfs Exp $";
+"$Id: relay.c,v 1.22 2001/06/14 20:59:20 dfs Exp $";
 
 #define _GNU_SOURCE 1 /* For SA_RESTART */
 
@@ -276,6 +276,13 @@ main(int argc, char *argv[])
 	}
     }
 
+#ifdef USE_LINUX_PACKET
+#ifndef HAVE_STRUCT_SOCKADDR_LL
+    fprintf(stderr, "The PPPoE relay does not work on Linux 2.0 kernels.\n");
+    exit(EXIT_FAILURE);
+#endif
+#endif
+
     /* Check that at least two interfaces were defined */
     if (NumInterfaces < 2) {
 	fprintf(stderr, "%s: Must define at least two interfaces\n",
@@ -356,6 +363,14 @@ addInterface(char const *ifname,
 	     int acOK)
 {
     PPPoEInterface *i;
+    int j;
+    for (j=0; j<NumInterfaces; j++) {
+	if (!strncmp(Interfaces[j].name, ifname, IFNAMSIZ)) {
+	    fprintf(stderr, "Interface %s specified more than once.\n", ifname);
+	    exit(EXIT_FAILURE);
+	}
+    }
+
     if (NumInterfaces >= MAX_INTERFACES) {
 	fprintf(stderr, "Too many interfaces (%d max)\n",
 		MAX_INTERFACES);
@@ -880,6 +895,11 @@ relayGotSessionPacket(PPPoEInterface const *iface)
 	return;
     }
 
+    /* Drop Ethernet frame padding */
+    if (size > ntohs(packet.length) + HDR_SIZE) {
+	size = ntohs(packet.length) + HDR_SIZE;
+    }
+
     /* We're in business!  Find the hash */
     sh = findSession(packet.ethHdr.h_source, packet.session);
     if (!sh) {
@@ -904,7 +924,7 @@ relayGotSessionPacket(PPPoEInterface const *iface)
 	    sh->peerMac[3], sh->peerMac[4], sh->peerMac[5],
 	    sh->interface->name, ntohs(sh->sesNum));
 #endif
-    sendPacket(sh->interface->sessionSock, &packet, size);
+    sendPacket(NULL, sh->interface->sessionSock, &packet, size);
 }
 
 /**********************************************************************
@@ -935,7 +955,7 @@ relayHandlePADT(PPPoEInterface const *iface,
     packet->session = sh->sesNum;
     memcpy(packet->ethHdr.h_source, sh->interface->mac, ETH_ALEN);
     memcpy(packet->ethHdr.h_dest, sh->peerMac, ETH_ALEN);
-    sendPacket(sh->interface->sessionSock, packet, size);
+    sendPacket(NULL, sh->interface->sessionSock, packet, size);
 
     /* Destroy the session */
     freeSession(ses, "Received PADT");
@@ -1030,7 +1050,7 @@ relayHandlePADI(PPPoEInterface const *iface,
 	if (iface == &Interfaces[i]) continue;
 	if (!Interfaces[i].acOK) continue;
 	memcpy(packet->ethHdr.h_source, Interfaces[i].mac, ETH_ALEN);
-	sendPacket(Interfaces[i].discoverySock, packet, size);
+	sendPacket(NULL, Interfaces[i].discoverySock, packet, size);
     }
     
 }
@@ -1148,7 +1168,7 @@ relayHandlePADO(PPPoEInterface const *iface,
     memcpy(packet->ethHdr.h_source, Interfaces[ifIndex].mac, ETH_ALEN);
 
     /* Send the PADO to the proper client */
-    sendPacket(Interfaces[ifIndex].discoverySock, packet, size);
+    sendPacket(NULL, Interfaces[ifIndex].discoverySock, packet, size);
 }
 
 /**********************************************************************
@@ -1264,7 +1284,7 @@ relayHandlePADR(PPPoEInterface const *iface,
     memcpy(packet->ethHdr.h_source, Interfaces[ifIndex].mac, ETH_ALEN);
 
     /* Send the PADR to the proper access concentrator */
-    sendPacket(Interfaces[ifIndex].discoverySock, packet, size);
+    sendPacket(NULL, Interfaces[ifIndex].discoverySock, packet, size);
 }
 
 /**********************************************************************
@@ -1422,7 +1442,7 @@ relayHandlePADS(PPPoEInterface const *iface,
     memcpy(packet->ethHdr.h_source, Interfaces[ifIndex].mac, ETH_ALEN);
 
     /* Send the PADS to the proper client */
-    sendPacket(Interfaces[ifIndex].discoverySock, packet, size);
+    sendPacket(NULL, Interfaces[ifIndex].discoverySock, packet, size);
 }
 
 /**********************************************************************
@@ -1469,9 +1489,9 @@ relaySendError(unsigned char code,
     if (addTag(&packet, &errTag) < 0) return;
     size = ntohs(packet.length) + HDR_SIZE;
     if (code == CODE_PADT) {
-	sendPacket(iface->discoverySock, &packet, size);
+	sendPacket(NULL, iface->discoverySock, &packet, size);
     } else {
-	sendPacket(iface->sessionSock, &packet, size);
+	sendPacket(NULL, iface->sessionSock, &packet, size);
     }
 }
 
